theory Finite_Runs
  imports Completion_Fairness
begin

locale kb_frun = kb +
  fixes R E n
  assumes R0: "R 0 = {}" and enc: "\<not> enc"
    and run: "\<And>i. i < n \<Longrightarrow> (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))"
begin

abbreviation Rinf ("R\<^sub>\<infinity>") where "R\<^sub>\<infinity> \<equiv> (\<Union>i\<le>n. R i)"
abbreviation Einf ("E\<^sub>\<infinity>") where "E\<^sub>\<infinity> \<equiv> (\<Union>i\<le>n. E i)"

definition "R\<^sub>\<omega> = R n"
definition "E\<^sub>\<omega> = E n"

lemma run_R_less:
  assumes "i \<le> n"
  shows "R i \<subseteq> {\<succ>}"
proof -
  { fix i assume "i \<le> n"
    then have "KB\<^sup>*\<^sup>* (E 0, R 0) (E i, R i)"
      using run by (induct i) (auto intro: rtranclp.rtrancl_into_rtrancl) }
  from KB_rtrancl_rules_subset_less [OF this] and R0
  show ?thesis using assms by auto
qed

lemma rstep_R_inf_less: "rstep R\<^sub>\<infinity> \<subseteq> {\<succ>}"
  using run_R_less by (blast intro: ctxt subst)

context
  assumes nonfail: "E n = {}"
begin

lemma E_i_subset_join_R_inf:
  assumes "i \<le> n"
  shows "E i \<subseteq> (rstep R\<^sub>\<infinity>)\<^sup>\<down>" (is "_ \<subseteq> ?R\<^sup>\<down>")
proof
  fix s t assume "(s, t) \<in> E i"
  then show "(s, t) \<in> ?R\<^sup>\<down>" using assms
  proof (induct "n - i" arbitrary: s t i)
    case 0
    then show ?case using nonfail by simp
  next
    case (Suc k)
    then have "k = n - (Suc i)" (is "k = n - ?j") and "Suc i \<le> n" and less: "i < n" by arith+
    note IH = Suc(1) [OF this(1) _ this(2)]
    from KB_E_subset [OF run, OF less, THEN subsetD, OF \<open>(s, t) \<in> E i\<close>]
    consider "(s, t) \<in> E ?j" | "(s, t) \<in> ((R ?j)\<^sup>\<leftrightarrow>)\<^sup>=" | "(s, t) \<in> rstep (R ?j) O E ?j" | "(s, t) \<in> E ?j O (rstep (R ?j))\<inverse>"
      by blast
    then show ?case
      using \<open>Suc i \<le> n\<close>
      apply (cases)
         apply (auto dest!: IH)
       apply (rule_tac b = y in rtrancl_join_join; blast)
      apply (rule_tac b = y in join_rtrancl_join; blast)
      done
  qed
qed

lemma rstep_enc [simp]: "rstep_enc (R i) = rstep (R i)"
  using enc unfolding rstep_enc_def by simp

lemma source_step_R_inf_subset:
  "source_step R\<^sub>\<infinity> l \<subseteq> (\<Union>t\<in>{t. l \<succeq> t}. slab R\<^sub>\<omega> t)\<^sup>\<leftrightarrow>\<^sup>*" (is "source_step ?R l \<subseteq> ?L l")
proof
  fix t u assume "(t, u) \<in> source_step R\<^sub>\<infinity> l"
  then obtain i where "i \<le> n" and "(t, u) \<in> R i" and [simp]: "l = t" by blast
  from this(2,1) have "(t, u) \<in> ?L t"
  proof (induct "n - i" arbitrary: t u i)
    case 0
    then show ?case unfolding R\<^sub>\<omega>_def by auto
  next
    case (Suc k)
    then have "k = n - (Suc i)" (is "k = n - ?j") and "Suc i \<le> n" and less: "i < n" by arith+
    note IH = Suc(1) [OF this(1) _ this(2)]
    then have IH': "\<And>t u. (t, u) \<in> rstep (R (Suc i)) \<Longrightarrow> (t, u) \<in> (\<Union>t\<in>{u. t \<succeq> u}. slab R\<^sub>\<omega> t)\<^sup>\<leftrightarrow>\<^sup>*"
      by (metis rstepE slab_conv_below_ctxt_subst)
    have "t \<succ> u" using \<open>(t, u) \<in> R i\<close> using run_R_less [of i] and less by auto
    from KB_R_subset [OF run, OF less, THEN subsetD, OF \<open>(t, u) \<in> R i\<close>, simplified]
    show ?case
      using \<open>Suc i \<le> n\<close> and \<open>t \<succ> u\<close> unfolding rstep_converse
      apply (cases)
       apply (auto  dest: IH IH' dest!: E_i_subset_join_R_inf [OF \<open>Suc i \<le> n\<close>, THEN subsetD])
      apply (frule IH')
       defer
       apply (frule IH)
      apply (frule IH')
       apply auto
      sorry
  qed
  then show "(t, u) \<in> ?L l" by simp
qed

lemma slab_R_inf_subset:
  "slab R\<^sub>\<infinity> s \<subseteq> (\<Union>t\<in>{t. s \<succeq> t}. slab R\<^sub>\<omega> t)\<^sup>\<leftrightarrow>\<^sup>*" (is "slab ?R s \<subseteq> ?L s")
proof
  fix t u assume "(t, u) \<in> slab ?R s"
  then obtain C \<sigma> l r where "(l, r) \<in> ?R"
    and [simp]: "s = t" "t = C\<langle>l \<cdot> \<sigma>\<rangle>" "u = C\<langle>r \<cdot> \<sigma>\<rangle>" by fast
  with source_step_R_inf_subset [of l]
  have "(l, r) \<in> ?L l" by blast
  from slab_conv_below_ctxt_subst [OF this] show "(t, u) \<in> ?L s" by simp
qed

lemma slab_R_inf_conv:
  assumes "(t, u) \<in> (\<Union>a\<in>{a. s \<succeq> a}. slab R\<^sub>\<infinity> a)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(t, u) \<in> (\<Union>a\<in>{a. s \<succeq> a}. slab R\<^sub>\<omega> a)\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  unfolding conversion_def
proof (induct)
  case (step u v)
  from step(2) obtain a where "s \<succeq> a" and "(u, v) \<in> (slab R\<^sub>\<infinity> a)\<^sup>\<leftrightarrow>" by blast
  with slab_R_inf_subset [of a]
  have "(u, v) \<in> (\<Union>x\<in>{x. a \<succeq> x}. slab R\<^sub>\<omega> x)\<^sup>\<leftrightarrow>\<^sup>*" by (auto simp: conversion_inv)
  with \<open>s \<succeq> a\<close> have "(u, v) \<in> (\<Union>x\<in>{x. s \<succeq> x}. slab R\<^sub>\<omega> x)\<^sup>\<leftrightarrow>\<^sup>*"
    using slab_conv_below_label by blast
  with step(3) show ?case by (auto simp: conversion_def)
qed simp

end

end

locale kb_frun_inf =
  kb_frun less for less :: "('a, 'b :: infinite) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool" (infix "\<succ>" 50)
begin

lemma finite_run:
  assumes nonfail: "E n = {}"
    and nabla: "\<And>s t u. (t, u) \<in> nabla (R n) s \<Longrightarrow> (t, u) \<in> (rstep (R n))\<^sup>\<down> \<union> (\<Union>i\<le>n. (rstep (E i))\<^sup>\<leftrightarrow>)"
  shows "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>* \<and> SN (rstep (R n)) \<and> CR (rstep (R n))"
proof -
  have R_less: "\<And>i. i \<le> n \<Longrightarrow> R i \<subseteq> {(x, y). x \<succ> y}"
    and "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>*"
    and SN: "SN (rstep (R n))"
    using finite_runD [of R, OF R0, of E n, OF assms(1)] and run by auto
  moreover
  have "CR (rstep (R n))"
  proof -
    interpret lab: ars_labeled_sn "slab (R n)" UNIV "op \<succ>"
      by (standard) (auto intro: SN_less)
    interpret ars_source_decreasing "op \<succ>" "rstep (R n)" UNIV
    proof
      fix s t u :: "('a, 'b) term" assume "(s, t) \<in> slab (R n) s" and "(s, u) \<in> slab (R n) s"
      then have "(t, u) \<in> nabla (R n) s ^^ 2"
        using SN_imp_variable_condition [OF SN] by (intro peak_imp_nabla2) auto
      moreover
      { fix v w assume *: "(v, w) \<in> nabla (R n) s" "v = t \<or> w = u"
        then have "s \<succ> v" and "s \<succ> w"
          using rsteps_subset_less [OF R_less [OF order_refl]]
          by (auto elim!: nablaE simp: rstep_rstep)
        consider "(v, w) \<in> (rstep (R n))\<^sup>\<down>" | "(v, w) \<in> (\<Union>i\<le>n. rstep (E i))\<^sup>\<leftrightarrow>"
          using nabla [OF *(1)] by blast
        then have "(v, w) \<in> (\<Union>v\<in>{v\<in>UNIV. s \<succ> v}. slab (R n) v)\<^sup>\<leftrightarrow>\<^sup>*"
        proof (cases)
          case 1
          then obtain x where "(v, x) \<in> (rstep (R n))\<^sup>*" and "(w, x) \<in> (rstep (R n))\<^sup>*" by blast
          then have "(v, x) \<in> (\<Union>a\<in>{a\<in>UNIV. v \<succeq> a}. slab (R n) a)\<^sup>\<leftrightarrow>\<^sup>*"
            and "(w, x) \<in> (\<Union>v\<in>{v\<in>UNIV. w \<succeq> v}. slab (R n) v)\<^sup>\<leftrightarrow>\<^sup>*"
            using rsteps_slabI [OF _ _ rstep_subset_less [OF R_less]] by auto
          then have "(v, x) \<in> (\<Union>a\<in>{a\<in>UNIV. s \<succ> a}. slab (R n) a)\<^sup>\<leftrightarrow>\<^sup>*"
            and "(x, w) \<in> (\<Union>a\<in>{a\<in>UNIV. s \<succ> a}. slab (R n) a)\<^sup>\<leftrightarrow>\<^sup>*"
            using \<open>s \<succ> v\<close> and \<open>s \<succ> w\<close>
            by (auto intro: slab_conv_less_label simp: conversion_inv)
          then show ?thesis by (auto simp: conversion_def)
        next
          let ?R = "\<Union>i\<le>n. R i"
          case 2
          with E_i_subset_join_R_inf [OF nonfail]
          have "(v, w) \<in> (rstep ?R)\<^sup>\<down>" by blast
          then obtain x where "(v, x) \<in> (rstep ?R)\<^sup>*" and "(w, x) \<in> (rstep ?R)\<^sup>*" by blast
          then have "(v, x) \<in> (\<Union>a\<in>{a\<in>UNIV. v \<succeq> a}. slab ?R a)\<^sup>\<leftrightarrow>\<^sup>*"
            and "(x, w) \<in> (\<Union>v\<in>{v\<in>UNIV. w \<succeq> v}. slab ?R v)\<^sup>\<leftrightarrow>\<^sup>*"
          using rsteps_slabI [OF _ _ rstep_R_inf_less] by (auto simp: conversion_inv)
          with slab_R_inf_conv [OF nonfail]
          have "(v, x) \<in> (\<Union>a\<in>{a\<in>UNIV. v \<succeq> a}. slab R\<^sub>\<omega> a)\<^sup>\<leftrightarrow>\<^sup>*"
            and "(x, w) \<in> (\<Union>a\<in>{a\<in>UNIV. w \<succeq> a}. slab R\<^sub>\<omega> a)\<^sup>\<leftrightarrow>\<^sup>*" by auto
          then have "(v, x) \<in> (\<Union>a\<in>{a\<in>UNIV. s \<succ> a}. slab R\<^sub>\<omega> a)\<^sup>\<leftrightarrow>\<^sup>*"
            and "(x, w) \<in> (\<Union>a\<in>{a\<in>UNIV. s \<succ> a}. slab R\<^sub>\<omega> a)\<^sup>\<leftrightarrow>\<^sup>*"
            using \<open>s \<succ> v\<close> and \<open>s \<succ> w\<close>
            by (auto intro: slab_conv_less_label simp: conversion_inv)
          then show ?thesis unfolding R\<^sub>\<omega>_def by (auto simp: conversion_def)
        qed }
      ultimately show "(t, u) \<in> (\<Union>v\<in>{v\<in>UNIV. s \<succ> v}. slab (R n) v)\<^sup>\<leftrightarrow>\<^sup>*"
        by (auto simp: conversion_def) (blast dest: rtrancl_trans)
    qed

    from CR show "CR (rstep (R n))" by (simp add: UN_source_step)
  qed
  ultimately show ?thesis by blast
qed

end

end
