theory Monads
  imports Main "~~/src/HOL/Library/Monad_Syntax"
begin
  
locale monad =
  fixes return :: "'a \<Rightarrow> 'b"
    and bind :: "'b \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b"
  assumes bind_assoc [simp]: "bind (bind m f) g = bind m (\<lambda>x. bind (f x) g)"
    and bind_return [simp]: "bind m return = m"
    and return_bind [simp]: "bind (return x) f = f x"
begin
  
adhoc_overloading Monad_Syntax.bind bind
end

datatype 'a exp =
  Const int
  | Var 'a
  | Add "'a exp" "'a exp" (infix "\<oplus>" 65)
  | Div "'a exp" "'a exp" (infix "\<oslash>" 65)

locale monad_fail = monad +
  fixes fail :: 'b
  assumes fail_bind [simp]: "fail \<bind> f = fail"

locale monad_fail_int =
  monad_fail return bind fail
  for return :: "int \<Rightarrow> 'b" and bind :: "'b \<Rightarrow> (int \<Rightarrow> 'b) \<Rightarrow> 'b" and fail :: "'b"
begin
  
primrec eval :: "('v \<Rightarrow> 'b) \<Rightarrow> 'v exp \<Rightarrow> 'b"
  where
    "eval E (Const i) = return i"
  | "eval E (Var x) = E x"
  | "eval E (e\<^sub>1 \<oplus> e\<^sub>2) = do { x \<leftarrow> e\<^sub>1; y \<leftarrow> e\<^sub>2; return (x + y) }"
    
  
end
end
